@@ -29,15 +29,19 @@ type Connector interface {
 	// connection - but a connector can choose to block this call until
 	// previously returned connections have been closed.
 	Connect(ctx context.Context) (net.Conn, error)
+
+	// If available, a unique identifier for the engine, otherwise empty string.
+	EngineID() string
 }
 
 type DriverOpts struct {
-	DaggerCloudToken string
-	GPUSupport       string
-	Module           string
-	Function         string
-	ExecCmd          []string
-	ClientID         string
+	DaggerCloudToken    string
+	GPUSupport          string
+	Module              string
+	Function            string
+	ExecCmd             []string
+	ClientID            string
+	CloudBasicAuthToken string
 }
 
 const (
