@@ -14,6 +14,7 @@ import (
 	"os"
 
 	"github.com/shurcooL/graphql"
+	"go.opentelemetry.io/otel/trace"
 	"golang.org/x/oauth2"
 
 	"github.com/dagger/dagger/engine"
@@ -23,13 +24,16 @@ import (
 var ErrNoOrg = errors.New("no org associated with this Engine")
 
 type Client struct {
-	u           *url.URL
-	g           *graphql.Client
-	h           *http.Client
-	engineToken string
+	u              *url.URL
+	g              *graphql.Client
+	h              *http.Client
+	basicAuthToken string
 }
 
-func NewClient(ctx context.Context) (*Client, error) {
+func NewClient(
+	ctx context.Context,
+	basicAuthToken string, // optional, if not set tries default sources
+) (*Client, error) {
 	api := "https://api.dagger.cloud"
 	if cloudURL := os.Getenv("DAGGER_CLOUD_URL"); cloudURL != "" {
 		api = cloudURL
@@ -42,27 +46,28 @@ func NewClient(ctx context.Context) (*Client, error) {
 
 	httpClient := &http.Client{}
 
-	// Always prefer oauth if available. If not and a DAGGER_CLOUD_TOKEN
-	// is set then use Basic auth
-	tokenSource, err := auth.TokenSource(ctx)
-	if err != nil {
-		if cloudToken := os.Getenv("DAGGER_CLOUD_TOKEN"); cloudToken != "" {
-			httpClient.Transport, err = auth.DaggerCloudTransport(ctx, cloudToken)
-			if err != nil {
-				return nil, err
-			}
-
-			return &Client{
-				u:           u,
-				h:           httpClient,
-				engineToken: cloudToken,
-			}, nil
+	if basicAuthToken == "" {
+		basicAuthToken = os.Getenv("DAGGER_CLOUD_TOKEN")
+	}
+	if basicAuthToken != "" {
+		httpClient.Transport, err = auth.DaggerCloudTransport(ctx, basicAuthToken)
+		if err != nil {
+			return nil, err
 		}
 
+		return &Client{
+			u:              u,
+			h:              httpClient,
+			basicAuthToken: basicAuthToken,
+		}, nil
+	}
+
+	// try oauth token
+	tokenSource, err := auth.TokenSource(ctx)
+	if err != nil {
 		return nil, err
 	}
 	httpClient = oauth2.NewClient(ctx, tokenSource)
-
 	return &Client{
 		u: u,
 		g: graphql.NewClient(u.JoinPath("/query").String(), httpClient),
@@ -103,6 +108,7 @@ type EngineRequest struct {
 	ExecCmd              []string `json:"exec_cmd,omitempty"`
 	ClientID             string   `json:"client_id,omitempty"`
 	MinimumEngineVersion string   `json:"minimum_engine_version,omitempty"`
+	TraceID              string   `json:"trace_id,omitempty"`
 }
 
 type EngineSpec struct {
@@ -114,6 +120,7 @@ type EngineSpec struct {
 	UserID         string                   `json:"user_id,omitempty"`
 	URL            string                   `json:"url,omitempty"`
 	CertSerialized *SerializableCertificate `json:"cert,omitempty"`
+	InstanceID     string                   `json:"instance_id,omitempty"`
 }
 
 func (es *EngineSpec) TLSCertificate() (*tls.Certificate, error) {
@@ -165,6 +172,7 @@ func (c *Client) Engine(ctx context.Context, req EngineRequest) (*EngineSpec, er
 	}
 
 	req.MinimumEngineVersion = engine.MinimumEngineVersion
+	req.TraceID = trace.SpanContextFromContext(ctx).TraceID().String()
 	engineSpec := &EngineSpec{
 		Image:         "registry.dagger.io/engine:" + tag,
 		EngineRequest: req,
@@ -179,7 +187,7 @@ func (c *Client) Engine(ctx context.Context, req EngineRequest) (*EngineSpec, er
 		return nil, fmt.Errorf("failed to generate a remote Engine request: %w", err)
 	}
 
-	if c.engineToken == "" {
+	if c.basicAuthToken == "" {
 		org, err := auth.CurrentOrg()
 		if err != nil {
 			return nil, ErrNoOrg
